from .. import BaseStandaloneRoutine
from pathlib import Path


class UnknownRoutine(BaseStandaloneRoutine):
    categories = ['Other']
    targets = []
    iconFile = Path(__file__).parent / "unknown.png"
    tooltip = "Unknown routine"

    def __init__(self, exp, name=''):
        BaseStandaloneRoutine.__init__(self, exp, name=name)
        self.type = 'UnknownStandaloneRoutine'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

from past.builtins import basestring
import numpy
import copy
from os import path
from psychopy import logging
from psychopy.constants import (STARTED, PLAYING, PAUSED, FINISHED, STOPPED,
                                NOT_STARTED, FOREVER)
from psychopy.tools.filetools import pathToString
from sys import platform
from .audioclip import AudioClip


if platform == 'win32':
    mediaLocation = "C:\\Windows\\Media"
elif platform == 'darwin':
    mediaLocation = "/System/Library/Sounds/"
elif platform.startswith("linux"):
    mediaLocation = "/usr/share/sounds"

stepsFromA = {
    'C': -9,
    'Csh': -8, 'C#': -8,
    'Dfl': -8, 'D♭': -8,
    'D': -7,
    'Dsh': -6, 'D#': -6,
    'Efl': -6, 'E♭': -6,
    'E': -5,
    'F': -4,
    'Fsh': -3, 'F#': -3,
    'Gfl': -3, 'G♭': -3,
    'G': -2,
    'Gsh': -1, 'G#': -1,
    'Afl': -1, 'A♭': -1,
    'A': +0,
    'Ash': +1, 'A#': +1,
    'Bfl': +1, 'B♭': +1,
    'B': +2,
    'Bsh': +2, 'B#': +2}
knownNoteNames = sorted(stepsFromA.keys())


def apodize(soundArray, sampleRate):
    """Apply a Hanning window (5ms) to reduce a sound's 'click' onset / offset
    """
    hwSize = int(min(sampleRate // 200, len(soundArray) // 15))
    hanningWindow = numpy.hanning(2 * hwSize + 1)
    soundArray = copy.copy(soundArray)
    soundArray[:hwSize] *= hanningWindow[:hwSize]
    soundArray[-hwSize:] *= hanningWindow[hwSize + 1:]
    return soundArray


class HammingWindow(object):
    def __init__(self, winSecs, soundSecs, sampleRate):
        """

        :param winSecs:
        :param soundSecs:
        :param sampleRate:
        """
        self.sampleRate = sampleRate
        self.winSecs = winSecs
        self.winSamples = int(round(sampleRate*winSecs))
        self.soundSecs = soundSecs
        self.soundSamples = int(round(sampleRate*soundSecs))
        self.startWindow = numpy.hanning(self.winSamples*2)[0:self.winSamples]
        self.endWindow = numpy.hanning(self.winSamples*2)[self.winSamples:]
        self.finalWinStart = self.soundSamples-self.winSamples

    def nextBlock(self, t, blockSize):
        """Returns a block to be multiplied with the current sound block or 1.0

        :param t: current position in time (secs)
        :param blockSize: block size for the sound needing the hanning window
        :return: numpy array of length blockSize
        """
        startSample = int(t*self.sampleRate)
        if startSample < self.winSamples:
            # we're in beginning hanning window (start of sound)
            # 2 options:
            #  - block is fully within window
            #  - block starts in window but ends after window
            block = numpy.ones(blockSize)
            winEndII = min(self.winSamples,  # if block goes beyond hann win
                           startSample+blockSize)  # if block shorter
            blockEndII = min(self.winSamples-startSample,  # if block beyond
                             blockSize)  # if block shorter
            block[0:blockEndII] = self.startWindow[startSample:winEndII]
        elif startSample >= self.soundSamples:
            block = None  # the sound has finished (shouldn't have got here!)
        elif startSample >= self.finalWinStart-blockSize:
            # we're in final hanning window (end of sound)
            # More complicated, with 3 options:
            #  - block starts before win
            #  - start/end during win
            #  - start during but end after win
            block = numpy.ones(blockSize)  # the initial flat part
            blockStartII = max(self.finalWinStart-startSample,
                         0)  # if block start inside window
            blockEndII = min(blockSize,  # if block ends in hann win
                             self.soundSamples-startSample)  # ends after snd
            winStartII = max(0,  # current block ends in win but starts before
                             startSample-self.finalWinStart)
            winEndII = min(self.winSamples,
                           startSample+blockSize-self.finalWinStart)
            block[blockStartII:blockEndII] = \
                self.endWindow[winStartII:winEndII]
        else:
            block = None  # we're in the middle of sound so no need for window
        if block is not None:
            block.shape = [len(block), 1]
        return block


class _SoundBase(object):
    """Base class for sound object, from one of many ways.
    """
    # Must be provided by class SoundPygame or SoundPyo:
    # def __init__()
    # def play(self, fromStart=True, **kwargs):
    # def stop(self, log=True):
    # def getDuration(self):
    # def getVolume(self):
    # def setVolume(self, newVol, log=True):
    # def _setSndFromFile(self, fileName):
    # def _setSndFromArray(self, thisArray):

    def setSound(self, value, secs=0.5, octave=4, hamming=True, log=True):
        """Set the sound to be played.

        Often this is not needed by the user - it is called implicitly during
        initialisation.

        Parameters
        ----------
        value : ArrayLike, AudioClip, int or str
            If it's a number between 37 and 32767 then a tone will be generated
            at that frequency in Hz. It could be a string for a note ('A',
            'Bfl', 'B', 'C', 'Csh'. ...). Then you may want to specify which
            octave.O r a string could represent a filename in the current
            location, or media location, or a full path combo. Or by giving an
            Nx2 numpy array of floats (-1:1) or
            :class:`~psychopy.sound.AudioClip` instance.
        secs : float
            Duration of a tone if a note name is to `value`.
        octave : int
            Is only relevant if the value is a note name. Middle octave of a
            piano is 4. Most computers won't output sounds in the bottom octave
            (1) and the top octave (8) is generally painful.
        hamming : bool
            To indicate if the sound should be apodized (i.e., the onset and
            offset smoothly ramped up from down to zero). The function apodize
            uses a Hanning window, but arguments named 'hamming' are preserved
            so that existing code is not broken by the change from Hamming to
            Hanning internally. Not applied to sounds from files.

        """
        # Re-init sound to ensure bad values will raise error during setting:
        self._snd = None

        # Coerces pathlib obj to string, else returns inputted value
        value = pathToString(value)
        try:
            # could be '440' meaning 440
            value = float(value)
        except (ValueError, TypeError):
            # this is a string that can't be a number, eg, a file or note
            pass
        else:
            # we've been asked for a particular Hz
            if value < 37 or value > 20000:
                msg = 'Sound: bad requested frequency %.0f'
                raise ValueError(msg % value)
            self._setSndFromFreq(value, secs, hamming=hamming)
        if isinstance(value, basestring):
            if value.capitalize() in knownNoteNames:
                self._setSndFromNote(value.capitalize(), secs, octave,
                                     hamming=hamming)
            else:
                # try finding a file
                self.fileName = None
                for filePath in ['', mediaLocation]:
                    p = path.join(filePath, value)
                    if path.isfile(p):
                        self.fileName = p
                        break
                    elif path.isfile(p + '.wav'):
                        self.fileName = p = p + '.wav'
                        break
                if self.fileName is None:
                    msg = "setSound: could not find a sound file named "
                    raise ValueError(msg + value)
                else:
                    self._setSndFromFile(p)
        elif isinstance(value, (list, numpy.ndarray,)):
            # create a sound from the input array/list
            self._setSndFromArray(numpy.array(value))
        elif isinstance(value, AudioClip):
            # from an audio clip object
            self._setSndFromArray(value.samples)
            self.sampleRate = value.sampleRateHz
        # did we succeed?
        if self._snd is None:
            pass  # raise ValueError, "Could not make a "+value+" sound"
        else:
            if log and self.autoLog:
                logging.exp("Set %s sound=%s" % (self.name, value), obj=self)
            self.status = NOT_STARTED

    def _setSndFromNote(self, thisNote, secs, octave, hamming=True):
        # note name -> freq -> sound
        freqA = 440.0
        thisOctave = octave - 4
        mult = 2.0**(stepsFromA[thisNote] / 12.)
        thisFreq = freqA * mult * 2.0 ** thisOctave
        self._setSndFromFreq(thisFreq, secs, hamming=hamming)

    def _setSndFromFreq(self, thisFreq, secs, hamming=True):
        # note freq -> array -> sound
        if secs < 0:
            # want infinite duration - create 1 sec sound and loop it
            secs = 10.0
            self.loops = -1
        if not self.sampleRate:
            self.sampleRate = self._getDefaultSampleRate()
        nSamples = int(secs * self.sampleRate)
        outArr = numpy.arange(0.0, 1.0, 1.0 / nSamples)
        outArr *= 2 * numpy.pi * thisFreq * secs
        outArr = numpy.sin(outArr)
        if hamming and nSamples > 30:
            outArr = apodize(outArr, self.sampleRate)
        self._setSndFromArray(outArr)

    def _getDefaultSampleRate(self):
        """For backends this might depend on what streams are open"""
        return 44100

    def getDuration(self):
        """Return the duration of the sound"""
        return self.duration

    def getVolume(self):
        """Returns the current volume of the sound (0.0 to 1.0, inclusive)
        """
        return self.volume

    def getLoops(self):
        """Returns the current requested loops value for the sound (int)
        """
        return self.loops

    def setVolume(self, newVol, log=True):
        """Sets the current volume of the sound (0.0 to 1.0, inclusive)
        """
        self.volume = min(1.0, max(0.0, newVol))
        self.needsUpdate = True
        if log and self.autoLog:
            logging.exp("Sound %s set volume %.3f" %
                        (self.name, self.volume), obj=self)

    def setLoops(self, newLoops, log=True):
        """Sets the current requested extra loops (int)"""
        self.loops = int(newLoops)
        self.needsUpdate = True
        if log and self.autoLog:
            logging.exp("Sound %s set loops %s" %
                        (self.name, self.loops), obj=self)



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).


from psychopy import visual
import gevent
from psychopy.iohub.util import convertCamelToSnake, updateSettings
from psychopy.iohub.devices import DeviceEvent, Computer
from psychopy.iohub.constants import EventConstants as EC
from psychopy.iohub.errors import print2err

currentTime = Computer.getTime


class MouseGazePsychopyCalibrationGraphics(object):
    IOHUB_HEARTBEAT_INTERVAL = 0.050
    CALIBRATION_POINT_LIST = [(0.5, 0.5), (0.1, 0.1), (0.9, 0.1), (0.9, 0.9), (0.1, 0.9)]

    _keyboard_key_index = EC.getClass(EC.KEYBOARD_RELEASE).CLASS_ATTRIBUTE_NAMES.index('key')

    def __init__(self, eyetrackerInterface, calibration_args):
        self._eyetracker = eyetrackerInterface
        self.screenSize = eyetrackerInterface._display_device.getPixelResolution()
        self.width = self.screenSize[0]
        self.height = self.screenSize[1]
        self._ioKeyboard = None
        self._msg_queue = []
        self._lastCalibrationOK = False
        self._device_config = self._eyetracker.getConfiguration()
        display = self._eyetracker._display_device

        updateSettings(self._device_config.get('calibration'), calibration_args)
        self._calibration_args = self._device_config.get('calibration')
        print2err("self._calibration_args:", self._calibration_args)
        unit_type = self.getCalibSetting('unit_type')
        if unit_type is None:
            unit_type = display.getCoordinateType()
            self._calibration_args['unit_type'] = unit_type
        color_type = self.getCalibSetting('color_type')
        if color_type is None:
            color_type = display.getColorSpace()
            self._calibration_args['color_type'] = color_type

        calibration_methods = dict(THREE_POINTS=3,
                                   FIVE_POINTS=5,
                                   NINE_POINTS=9,
                                   THIRTEEN_POINTS=13)

        cal_type = self.getCalibSetting('type')

        if cal_type in calibration_methods:
            num_points = calibration_methods[cal_type]

            if num_points == 3:
                MouseGazePsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.1),
                                                                               (0.1, 0.9),
                                                                               (0.9, 0.9)]
            elif num_points == 5:
                MouseGazePsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.5),
                                                                               (0.1, 0.1),
                                                                               (0.9, 0.1),
                                                                               (0.9, 0.9),
                                                                               (0.1, 0.9)]
            elif num_points == 9:
                MouseGazePsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.5),
                                                                               (0.1, 0.5),
                                                                               (0.9, 0.5),
                                                                               (0.1, 0.1),
                                                                               (0.5, 0.1),
                                                                               (0.9, 0.1),
                                                                               (0.9, 0.9),
                                                                               (0.5, 0.9),
                                                                               (0.1, 0.9)]
            elif num_points == 13:
                MouseGazePsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.5),
                                                                               (0.1, 0.5),
                                                                               (0.9, 0.5),
                                                                               (0.1, 0.1),
                                                                               (0.5, 0.1),
                                                                               (0.9, 0.1),
                                                                               (0.9, 0.9),
                                                                               (0.5, 0.9),
                                                                               (0.1, 0.9),
                                                                               (0.25, 0.25),
                                                                               (0.25, 0.75),
                                                                               (0.75, 0.75),
                                                                               (0.75, 0.25)
                                                                               ]

        self.window = visual.Window(
            self.screenSize,
            monitor=display.getPsychopyMonitorName(),
            units=unit_type,
            fullscr=True,
            allowGUI=False,
            screen=display.getIndex(),
            color=self.getCalibSetting(['screen_background_color']),
            colorSpace=color_type)
        self.window.flip(clearBuffer=True)

        self._createStim()
        self._registerEventMonitors()
        self._lastMsgPumpTime = currentTime()

        self.clearAllEventBuffers()

    def getCalibSetting(self, setting):
        if isinstance(setting, str):
            setting = [setting, ]
        calibration_args = self._calibration_args
        if setting:
            for s in setting[:-1]:
                calibration_args = calibration_args.get(s)
            return calibration_args.get(setting[-1])

    def clearAllEventBuffers(self):
        self._eyetracker._iohub_server.eventBuffer.clear()
        for d in self._eyetracker._iohub_server.devices:
            d.clearEvents()

    def _registerEventMonitors(self):
        kbDevice = None
        if self._eyetracker._iohub_server:
            for dev in self._eyetracker._iohub_server.devices:
                if dev.__class__.__name__ == 'Keyboard':
                    kbDevice = dev

        if kbDevice:
            eventIDs = []
            for event_class_name in kbDevice.__class__.EVENT_CLASS_NAMES:
                eventIDs.append(getattr(EC, convertCamelToSnake(event_class_name[:-5], False)))

            self._ioKeyboard = kbDevice
            self._ioKeyboard._addEventListener(self, eventIDs)
        else:
            print2err(
                'Warning: GazePoint Cal GFX could not connect to Keyboard device for events.')

    def _unregisterEventMonitors(self):
        if self._ioKeyboard:
            self._ioKeyboard._removeEventListener(self)

    def _handleEvent(self, event):
        event_type_index = DeviceEvent.EVENT_TYPE_ID_INDEX
        if event[event_type_index] == EC.KEYBOARD_PRESS:
            ek = event[self._keyboard_key_index]
            if isinstance(ek, bytes):
                ek = ek.decode('utf-8')
            if ek == ' ':
                self._msg_queue.append('SPACE_KEY_ACTION')
                self.clearAllEventBuffers()
            elif ek == 'escape':
                self._msg_queue.append('QUIT')
                self.clearAllEventBuffers()

    def MsgPump(self):
        # keep the psychopy window happy ;)
        if currentTime() - self._lastMsgPumpTime > self.IOHUB_HEARTBEAT_INTERVAL:
            # try to keep ioHub from being blocked. ;(
            if self._eyetracker._iohub_server:
                for dm in self._eyetracker._iohub_server.deviceMonitors:
                    dm.device._poll()
                self._eyetracker._iohub_server.processDeviceEvents()
            self._lastMsgPumpTime = currentTime()

    def getNextMsg(self):
        if len(self._msg_queue) > 0:
            msg = self._msg_queue[0]
            self._msg_queue = self._msg_queue[1:]
            return msg

    def _createStim(self):
        """
            outer_diameter: 35
            outer_stroke_width: 5
            outer_fill_color: [255,255,255]
            outer_line_color: [255,255,255]
            inner_diameter: 5
            inner_stroke_width: 0
            inner_color: [0,0,0]
            inner_fill_color: [0,0,0]
            inner_line_color: [0,0,0]
            calibration_prefs=self._eyetracker.getConfiguration()['calibration']['target_attributes']
        """
        color_type = self.getCalibSetting('color_type')
        unit_type = self.getCalibSetting('unit_type')

        lwidth = self.getCalibSetting(['target_attributes', 'outer_stroke_width'])
        radius = self.getCalibSetting(['target_attributes', 'outer_diameter']) / 2.0
        fcolor = self.getCalibSetting(['target_attributes', 'outer_fill_color'])
        lcolor = self.getCalibSetting(['target_attributes', 'outer_line_color'])
        self.calibrationPointOUTER = visual.Circle(self.window, pos=(0, 0), name='CP_OUTER',
                                                   radius=radius, lineWidth=lwidth,
                                                   fillColor=fcolor, lineColor=lcolor,
                                                   opacity=1.0, interpolate=False,
                                                   edges=64, units=unit_type, colorSpace=color_type)

        lwidth = self.getCalibSetting(['target_attributes', 'inner_stroke_width'])
        radius = self.getCalibSetting(['target_attributes', 'inner_diameter']) / 2.0
        fcolor = self.getCalibSetting(['target_attributes', 'inner_fill_color'])
        lcolor = self.getCalibSetting(['target_attributes', 'inner_line_color'])
        self.calibrationPointINNER = visual.Circle(self.window, pos=(0, 0), name='CP_INNER',
                                                   radius=radius, lineWidth=lwidth,
                                                   fillColor=fcolor, lineColor=lcolor,
                                                   opacity=1.0, interpolate=False,
                                                   edges=64, units=unit_type, colorSpace=color_type)

        instuction_text = 'Press SPACE to Start Calibration; ESCAPE to Exit.'
        self.textLineStim = visual.TextStim(self.window, text=instuction_text,
                                            pos=(0, 0), height=36,
                                            color=(0, 0, 0), colorSpace='rgb255',
                                            units='pix', wrapWidth=self.width * 0.9)

    def runCalibration(self):
        """Run calibration sequence
        """
        instuction_text = 'Press SPACE to Start Calibration.'
        self.showSystemSetupMessageScreen(instuction_text)

        target_delay = self.getCalibSetting('target_delay')
        target_duration = self.getCalibSetting('target_duration')
        auto_pace = self.getCalibSetting('auto_pace')
        cal_target_list = self.CALIBRATION_POINT_LIST
        randomize_points = self.getCalibSetting('randomize')
        print2err('randomize:', randomize_points)
        if randomize_points is True:
            # Randomize all but first target position.
            cal_target_list = self.CALIBRATION_POINT_LIST[1:]
            import random
            random.seed(None)
            random.shuffle(cal_target_list)
            cal_target_list.insert(0, self.CALIBRATION_POINT_LIST[0])

        left, top, right, bottom = self._eyetracker._display_device.getCoordBounds()
        w, h = right - left, top - bottom

        self.clearCalibrationWindow()

        i = 0
        abort_calibration = False
        for pt in cal_target_list:
            if abort_calibration:
                break
            # Convert normalized positions to psychopy window unit positions
            # by using iohub display/window getCoordBounds.
            x, y = left + w * pt[0], bottom + h * (1.0 - pt[1])
            start_time = currentTime()

            self.clearAllEventBuffers()

            # Target animate / delay
            animate_enable = self.getCalibSetting(['target_attributes', 'animate', 'enable'])
            animate_expansion_ratio = self.getCalibSetting(['target_attributes', 'animate', 'expansion_ratio'])
            animate_contract_only = self.getCalibSetting(['target_attributes', 'animate', 'contract_only'])
            while currentTime()-start_time <= target_delay:
                if animate_enable and i > 0:
                    t = (currentTime()-start_time) / target_delay
                    v1 = cal_target_list[i-1]
                    v2 = pt
                    t = 60.0 * ((1.0 / 10.0) * t ** 5 - (1.0 / 4.0) * t ** 4 + (1.0 / 6.0) * t ** 3)
                    mx, my = ((1.0 - t) * v1[0] + t * v2[0], (1.0 - t) * v1[1] + t * v2[1])
                    moveTo = left + w * mx, bottom + h * (1.0 - my)
                    self.drawCalibrationTarget(moveTo, reset=False)
                else:
                    self.drawCalibrationTarget((x, y), False)

            gevent.sleep(0.001)
            self.MsgPump()
            msg = self.getNextMsg()
            if msg == 'QUIT':
                abort_calibration = True
                break

            # Target expand / contract phase
            self.drawCalibrationTarget((x, y))
            start_time = currentTime()
            outer_diameter = self.getCalibSetting(['target_attributes', 'outer_diameter'])
            inner_diameter = self.getCalibSetting(['target_attributes', 'inner_diameter'])
            while currentTime()-start_time <= target_duration:
                elapsed_time = currentTime()-start_time
                d = t = None
                if animate_contract_only:
                    # Change target size from outer diameter to inner diameter over target_duration seconds.
                    t = elapsed_time / target_duration
                    d = outer_diameter - t * (outer_diameter - inner_diameter)
                    self.calibrationPointOUTER.radius = d / 2
                    self.calibrationPointOUTER.draw()
                    self.calibrationPointINNER.draw()
                    self.window.flip(clearBuffer=True)
                elif animate_expansion_ratio not in [1, 1.0]:
                    if elapsed_time <= target_duration/2:
                        # In expand phase
                        t = elapsed_time / (target_duration/2)
                        d = outer_diameter + t * (outer_diameter*animate_expansion_ratio - outer_diameter)
                    else:
                        # In contract phase
                        t = (elapsed_time-target_duration/2) / (target_duration/2)
                        d = outer_diameter*animate_expansion_ratio - t * (outer_diameter*animate_expansion_ratio - inner_diameter)
                if d:
                    self.calibrationPointOUTER.radius = d / 2
                    self.calibrationPointOUTER.draw()
                    self.calibrationPointINNER.draw()
                    self.window.flip(clearBuffer=True)

            if auto_pace is False:
                while 1:
                    gevent.sleep(0.001)
                    self.MsgPump()
                    msg = self.getNextMsg()
                    if msg == 'SPACE_KEY_ACTION':
                        break
                    elif msg == 'QUIT':
                        abort_calibration = True
                        break

            gevent.sleep(0.001)
            self.MsgPump()
            msg = self.getNextMsg()
            while msg:
                if msg == 'QUIT':
                    abort_calibration = True
                    break
                gevent.sleep(0.001)
                self.MsgPump()
                msg = self.getNextMsg()

            self.clearCalibrationWindow()
            self.clearAllEventBuffers()
            i += 1

        if abort_calibration is False:
            instuction_text = "Calibration Complete. Press 'SPACE' key to continue."
            self.showSystemSetupMessageScreen(instuction_text)
            return True
        return False

    def clearCalibrationWindow(self):
        self.window.flip(clearBuffer=True)

    def showSystemSetupMessageScreen(self, text_msg='Press SPACE to Start Calibration; ESCAPE to Exit.'):

        self.clearAllEventBuffers()

        while True:
            self.textLineStim.setText(text_msg)
            self.textLineStim.draw()
            self.window.flip()

            msg = self.getNextMsg()
            if msg == 'SPACE_KEY_ACTION':
                self.clearAllEventBuffers()
                return True
            elif msg == 'QUIT':
                self.clearAllEventBuffers()
                return False
            self.MsgPump()
            gevent.sleep(0.001)

    def drawDefaultTarget(self):
        self.calibrationPointOUTER.radius = self.getCalibSetting(['target_attributes', 'outer_diameter']) / 2.0
        self.calibrationPointOUTER.setLineColor(self.getCalibSetting(['target_attributes', 'outer_line_color']))
        self.calibrationPointOUTER.setFillColor(self.getCalibSetting(['target_attributes', 'outer_fill_color']))
        self.calibrationPointOUTER.lineWidth = int(self.getCalibSetting(['target_attributes', 'outer_stroke_width']))
        self.calibrationPointINNER.radius = self.getCalibSetting(['target_attributes', 'inner_diameter']) / 2.0
        self.calibrationPointINNER.setLineColor(self.getCalibSetting(['target_attributes', 'inner_line_color']))
        self.calibrationPointINNER.setFillColor(self.getCalibSetting(['target_attributes', 'inner_fill_color']))
        self.calibrationPointINNER.lineWidth = int(self.getCalibSetting(['target_attributes', 'inner_stroke_width']))

        self.calibrationPointOUTER.draw()
        self.calibrationPointINNER.draw()
        return self.window.flip(clearBuffer=True)

    def drawCalibrationTarget(self, tp, reset=True):
        self.calibrationPointOUTER.setPos(tp)
        self.calibrationPointINNER.setPos(tp)
        if reset:
            return self.drawDefaultTarget()
        else:
            self.calibrationPointOUTER.draw()
            self.calibrationPointINNER.draw()
            return self.window.flip(clearBuffer=True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Simple iohub eye tracker device demo.
Select which tracker to use by setting the TRACKER variable below.
"""

from __future__ import absolute_import, division, print_function
from psychopy import core, visual
from psychopy.iohub import launchHubServer
from psychopy.iohub.util import hideWindow, showWindow

# Eye tracker to use ('mouse', 'eyelink', 'gazepoint', or 'tobii')
TRACKER = 'mouse'

eyetracker_config = dict(name='tracker')
devices_config = {}
if TRACKER == 'mouse':
    devices_config['eyetracker.hw.mouse.EyeTracker'] = eyetracker_config
elif TRACKER == 'eyelink':
    eyetracker_config['model_name'] = 'EYELINK 1000 DESKTOP'
    eyetracker_config['runtime_settings'] = dict(sampling_rate=1000, track_eyes='RIGHT')
    devices_config['eyetracker.hw.sr_research.eyelink.EyeTracker'] = eyetracker_config
elif TRACKER == 'gazepoint':
    devices_config['eyetracker.hw.gazepoint.gp3.EyeTracker'] = eyetracker_config
elif TRACKER == 'tobii':
    devices_config['eyetracker.hw.tobii.EyeTracker'] = eyetracker_config
else:
    print("{} is not a valid TRACKER name; please use 'mouse', 'eyelink', 'gazepoint', or 'tobii'.".format(TRACKER))
    core.quit()

# Number if 'trials' to run in demo
TRIAL_COUNT = 2
# Maximum trial time / time timeout
T_MAX = 60.0
win = visual.Window((1920, 1080),
                    units='pix',
                    fullscr=True,
                    allowGUI=False,
                    colorSpace='rgb255',
                    monitor='55w_60dist',
                    color=[128, 128, 128]
                    )

win.setMouseVisible(False)
text_stim = visual.TextStim(win, text="Start of Experiment",
                            pos=[0, 0], height=24,
                            color='black', units='pix', colorSpace='named',
                            wrapWidth=win.size[0] * .9)

text_stim.draw()
win.flip()

io = launchHubServer(window=win, **devices_config)

# Get some iohub devices for future access.
keyboard = io.getDevice('keyboard')
tracker = io.getDevice('tracker')

# Minimize the PsychoPy window if needed
hideWindow(win)
# Display calibration gfx window and run calibration.
result = tracker.runSetupProcedure()
print("Calibration returned: ", result)
# Maximize the PsychoPy window if needed
showWindow(win)

gaze_ok_region = visual.Circle(win, lineColor='black', radius=300, units='pix', colorSpace='named')

gaze_dot = visual.GratingStim(win, tex=None, mask='gauss', pos=(0, 0),
                              size=(40, 40), color='green', colorSpace='named', units='pix')

text_stim_str = 'Eye Position: %.2f, %.2f. In Region: %s\n'
text_stim_str += 'Press space key to start next trial.'
missing_gpos_str = 'Eye Position: MISSING. In Region: No\n'
missing_gpos_str += 'Press space key to start next trial.'
text_stim.setText(text_stim_str)

# Run Trials.....
t = 0
while t < TRIAL_COUNT:
    io.clearEvents()
    tracker.setRecordingState(True)
    run_trial = True
    tstart_time = core.getTime()
    while run_trial is True:
        # Get the latest gaze position in display coord space.
        gpos = tracker.getLastGazePosition()
        # Update stim based on gaze position
        valid_gaze_pos = isinstance(gpos, (tuple, list))
        gaze_in_region = valid_gaze_pos and gaze_ok_region.contains(gpos)
        if valid_gaze_pos:
            # If we have a gaze position from the tracker, update gc stim and text stim.
            if gaze_in_region:
                gaze_in_region = 'Yes'
            else:
                gaze_in_region = 'No'
            text_stim.text = text_stim_str % (gpos[0], gpos[1], gaze_in_region)

            gaze_dot.setPos(gpos)
        else:
            # Otherwise just update text stim
            text_stim.text = missing_gpos_str

        # Redraw stim
        gaze_ok_region.draw()
        text_stim.draw()
        if valid_gaze_pos:
            gaze_dot.draw()

        # Display updated stim on screen.
        flip_time = win.flip()

        # Check any new keyboard char events for a space key.
        # If one is found, set the trial end variable.
        #
        if keyboard.getPresses(keys=' '):
            run_trial = False
        elif core.getTime()-tstart_time > T_MAX:
            run_trial = False
    win.flip()
    # Current Trial is Done
    # Stop eye data recording
    tracker.setRecordingState(False)
    t += 1

# All Trials are done
# End experiment
win.close()
tracker.setConnectionState(False)

io.quit()
core.quit()



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
from .eyetracker import EyeTracker
from psychopy.iohub.devices.eyetracker import (MonocularEyeSampleEvent, FixationStartEvent, FixationEndEvent,
                                               SaccadeEndEvent, SaccadeStartEvent, BlinkEndEvent, BlinkStartEvent)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Various string functions for working with strings.
#

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import re

__all__ = ["prettyname"]


def prettyname(name, wrap=False):
    """Convert a camelCase, TitleCase or underscore_delineated title to Full Title Case"""
    # Replace _ with space
    name = name.replace("_", " ")
    # Put a space before any capital letter, apart from at the beginning, or already after a space
    name = name[0] + re.sub('(?<![ -.])([A-Z])', r' \1', name[1:])
    # Capitalise first letter of each word
    name = name.title()
    # Treat the word "PsychoPy" as a special case
    name = name.replace("Psycho Py", "PsychoPy")
    # Split into multiple lines if wrap is requested
    if wrap:
        sentence = []
        letter = 0
        # Iterate through each word
        for word in name.split(" "):
            # Count its letters
            letter += len(word)
            if letter > wrap >= len(word):
                # If this brings the current letters this line to more than the wrap limit, insert a line break
                sentence.append("\n")
                letter = 0
            # Insert word
            sentence.append(word)
        # Recombine name
        name = " ".join(sentence)
        # Remove spaces after line
        name.replace(" \n ", "\n")

    return name


# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
"""iohub wintab util objects / functions for stylus, position traces,
and validation process psychopy graphics.
"""
from __future__ import division, absolute_import

import math
from collections import OrderedDict

import numpy as np

from psychopy import visual, core
from psychopy.visual.basevisual import MinimalStim

class PenPositionStim(MinimalStim):
    """Draws the current pen x,y position with graphics that represent the
    pressure, z axis, and tilt data for the wintab sample used."""
    def __init__(self, win, min_opacity=0.0, hover_color=(255,0,0), 
                 touching_color=(0,255,0), tiltline_color=(255,255, 0),
                 tiltline_width=2,
                 min_size=0.033, size_range=0.1666, tiltline_scalar=1.0,
                 name=None, autoLog=None, depth=-10000, colorSpace='rgb255'):
        self.win = win
        self.depth = depth
        super(PenPositionStim, self).__init__(name, autoLog)

        # Pen Hovering Related

        # Opaticy is changed based on pen's z axis if data for z axis
        # is available. Opacity of min_opacity is used when pen is at the
        # furthest hover distance (z value) supported by the device.
        # Opacity of 1.0 is used when z value == 0, meaning pen is touching
        # digitizer surface.
        self.min_opacity = min_opacity
        # If z axis is supported, hover_color specifies the color of the pen
        # position dot when z val > 0.
        self.hover_color = hover_color

        # Pen Pressure Related

        # Smallest radius (in norm units) that the pen position gaussian blob
        # will have, which occurs when pen pressure value is 0
        self.min_size = min_size
        # As pen pressure value increases, so does position gaussian blob
        # radius (in norm units). Max radius is reached when pressure is
        # at max device pressure value, and is equal to min_size+size_range
        self.size_range = size_range
        # Color of pen position blob when pressure > 0.
        self.touching_color = touching_color

        # Pen tilt Related

        # Color of line graphic used to represent the pens tilt relative to
        # the digitizer surface.
        self.tiltline_color = tiltline_color
        self.tiltline_width = tiltline_width
        self.tiltline_scalar = tiltline_scalar
        # Create a Gausian blob stim to use for pen position graphic
        self.pen_guass = visual.PatchStim(win, units='norm', tex='none',
                                          mask='gauss', pos=(0, 0), 
                                          colorSpace='rgb255',
                                          size=(self.min_size,self.min_size),
                                          color=self.hover_color,
                                          autoLog=False,
                                          opacity=0.0)

        # Create a line stim to use for pen position graphic
        self.pen_tilt_line = visual.Line(win, units='norm', start=[0, 0],
                                         lineWidth=self.tiltline_width,
                                         end=[0, 0],
                                         colorSpace='rgb255',
                                         lineColor=self.tiltline_color,
                                         opacity=0.0)
        #self.pen_tilt_line.opacity=0.0

    def updateFromEvent(self, evt):
        """Update the pen position and tilt graphics based on the data from
        a wintab sample event.

        :param evt: iohub wintab sample event
        :return:
        """
        # update the pen position stim based on
        # the last tablet event's data
        if evt.pressure > 0:
            # pen is touching tablet surface
            self.pen_guass.color = self.touching_color
        else:
            # pen is hovering just above tablet surface
            self.pen_guass.color = self.hover_color

        if evt.device.axis['pressure']['supported']:
            # change size of pen position blob based on samples pressure
            # value
            pnorm = evt.pressure / evt.device.axis['pressure']['range']
            self.pen_guass.size = self.min_size + pnorm * self.size_range

        # set the position of the gauss blob to be the pen x,y value converted
        # to norm screen coords.
        self.pen_guass.pos = evt.getNormPos()

        # if supported, update all graphics opacity based on the samples z value
        # otherwise opacity is always 1.0
        if evt.device.axis['z']['supported']:
            z = evt.device.axis['z']['range'] - evt.z
            znorm = z / evt.device.axis['z']['range']
            sopacity = self.min_opacity + znorm * (1.0 - self.min_opacity)
            self.pen_guass.opacity = self.pen_tilt_line.opacity = sopacity
        else:
            self.pen_guass.opacity = self.pen_tilt_line.opacity = 1.0

        # Change the tilt line start position to == pen position
        self.pen_tilt_line.start = self.pen_guass.pos

        # Change the tilt line end position based on samples tilt value
        # If tilt is not supported, it will always return 0,0
        # so no line is drawn.
        t1, t2 = evt.tilt
        pen_tilt_xy = 0, 0
        if t1 != t2 != 0:
            pen_tilt_xy = t1 * math.sin(t2), t1 * math.cos(t2)

        pen_pos = self.pen_guass.pos
        tiltend = (pen_pos[0] + pen_tilt_xy[0]*self.tiltline_scalar, 
                   pen_pos[1] + pen_tilt_xy[1]*self.tiltline_scalar)
        self.pen_tilt_line.end = tiltend

    def draw(self):
        """Draw the PenPositionStim to the opengl back buffer. This needs
        to be called prior to calling win.flip() for the stim is to be
        displayed.

        :return: None

        """
        self.pen_guass.draw()
        self.pen_tilt_line.draw()

    def clear(self):
        """Hide the graphics on the screen, even if they are drawn, by
        setting opacity to 0.

        :return: None

        """
        self.pen_guass.opacity = 0.0
        self.pen_tilt_line.opacity = 0.0

    def __del__(self):
        self.win = None

class PenTracesStim(MinimalStim):
    """Graphics representing where the pen has been moved on the digitizer
    surface. Positions where sample pressure > 0 are included.

    Implemented as a list of visual.ShapeStim, each representing a
    single pen trace/segment (series on pen samples with pressure >
    0). For improved performance, a single pen trace can have
    max_trace_len points before a new ShapeStim is created and made
    the 'current' pen trace'.
    """
    def __init__( self, win, lineWidth=2, lineColor=(0, 0, 0), opacity=1.0,
                 maxlen=256, name=None, autoLog=None, depth=-1000):
        self.depth = depth
        self.win = win
        super(PenTracesStim, self).__init__(name, autoLog)
        # A single pen trace can have at most max_trace_len points.
        self.max_trace_len = maxlen
        # The list of ShapeStim representing pen traces
        self.pentracestim = []
        # The ShapeStim state new / upcoming position points will be added to.
        self.current_pentrace = None
        # A list representation of the current_pentrace.vertices
        self.current_points = []
        # The last pen position added to a pen trace.
        self.last_pos = [0, 0]
        
        self.lineWidth=lineWidth
        self.lineColor=lineColor
        self.opacity=opacity


    @property
    def traces(self):
        """List of np arrays, each np array is the set of vertices for one
        pen trace.

        :return: list

        """
        return [pts.vertices for pts in self.pentracestim]

    def updateFromEvents(self, sample_events):
        """
        Update the stim graphics based on 0 - n pen sample events.
        :param sample_events:
        :return: None
        """
        for pevt in sample_events:
            if 'FIRST_ENTER' in pevt.status:
                self.end()
            if pevt.pressure > 0:
                lpx, lpy = self.last_pos
                px, py = pevt.getPixPos(self.win)
                if lpx != px or lpy != py:
                    if len(self.current_points) >= self.max_trace_len:
                        self.end()
                        self.append((lpx, lpy))
                    self.last_pos = (px, py)
                    self.append(self.last_pos)
            else:
                self.end()

    def draw(self):
        """Draws each pen trace ShapeStim to the opengl back buffer. This
        method must be called prior to calling win.flip() if it is to
        appear on the screen.

        :return: None
        """
        for pts in self.pentracestim:
            pts.draw()

    def start(self, first_point):
        """Start a new pen trace, by creating a new ShapeStim, adding it to
        the pentracestim list, and making it the current_pentrace.

        :param first_point: the first point in the ShapStim being craeted.
        :return: None
        """
        self.end()
        self.current_points.append(first_point)
        self.current_pentrace = visual.ShapeStim(self.win,
                                                 units='pix',
                                                 lineWidth=self.lineWidth,
                                                 color=self.lineColor,
                                                 lineColorSpace='rgb255',
                                                 vertices=self.current_points,
                                                 closeShape=False,
                                                 pos=(0, 0),
                                                 size=1,
                                                 ori=0.0,
                                                 opacity=self.opacity,
                                                 interpolate=True)
        self.pentracestim.append(self.current_pentrace)

    def end(self):
        """Stop using the current_pentrace ShapeStim. Next time a pen
        sample position is added to the PenTracesStim instance, a new
        ShapeStim will created and added to the pentracestim list.

        :return: None
        """
        self.current_pentrace = None
        self.current_points = []
        self.last_pos = [0, 0]

    def append(self, pos):
        """Add a pen position (in pix coords) to the current_pentrace
        ShapeStim vertices.

        :param pos: (x,y) tuple
        :return: None
        """
        if self.current_pentrace is None:
            self.start(pos)
        else:
            self.current_points.append(pos)
            self.current_pentrace.vertices = self.current_points

    def clear(self):
        """Remove all ShapStim being used. Next time this stim is drawn, no
        pen traces will exist.

        :return:
        """
        self.end()
        del self.pentracestim[:]

    def __del__(self):
        self.clear()
        self.win = None

#
# Pen position validation process code
#

class ScreenPositionValidation(object):
    NUM_VALID_SAMPLES_PER_TARG = 100
    TARGET_TIMEOUT = 10.0

    def __init__(self, win, io, target_stim=None, pos_grid=None,
                 display_pen_pos=True, force_quit=True, intro_title=None,
                 intro_text1=None, intro_text2=None, intro_target_pos=None):
        """ScreenPositionValidation is used to perform a pen position
        accuracy test for an iohub wintab device.

        :param win: psychopy Window instance to ude for the validation graphics
        :param io: iohub connection instance
        :param target_stim: None to use default, or  psychopy.iohub.util.targetpositionsequence.TargetStim instance
        :param pos_grid: None to use default, or  psychopy.iohub.util.targetpositionsequence.PositionGrid instance
        :param display_pen_pos: True to add calculated pen position graphic
        :param force_quit: Not Used
        :param intro_title: None to use default, str or unicode to set the text used for the introduction screen title, or an instance of psychopy.visual.TextStim
        :param intro_text1: None to use default, str or unicode to set the text used for the introduction text part 1, or an instance of psychopy.visual.TextStim
        :param intro_text2: None to use default, str or unicode to set the text used for the introduction text part 2, or an instance of psychopy.visual.TextStim
        :param intro_target_pos: None to use default, or (x,y) position to place the target graphic on the introduction screen. (x,y) position must be specified in 'norm' coordinate space.
        :return:
        """

        from psychopy.iohub.util.targetpositionsequence import TargetStim, PositionGrid

        self.win = win
        self.io = io
        self._lastPenSample = None
        self._targetStim = target_stim
        self._positionGrid = pos_grid
        self._forceQuit = force_quit
        self._displayPenPosition = display_pen_pos

        # IntroScreen Graphics
        intro_graphics = self._introScreenGraphics = OrderedDict()

        # Title Text
        title_stim = visual.TextStim(self.win, units='norm',
                                     pos=(0, .9),
                                     height=0.1,
                                     text='Pen Position Validation')
        if isinstance(intro_title, str):
            title_stim.setText(intro_title)
        elif isinstance(intro_title, visual.TextStim):
            title_stim = intro_title
        intro_graphics['title'] = title_stim

        # Intro Text part 1
        text1_stim = visual.TextStim(self.win, units='norm',
                                     pos=(0, .65),
                                     height=0.05,
                                     text='On the following screen, '
                                     'press the pen on the target '
                                     'graphic when it appears, '
                                     'as accurately as '
                                     'possible, until the target '
                                     'moves to a different '
                                     'location. Then press at the '
                                     'next target location. '
                                     'Hold the stylus in exactly '
                                     'the same way as you would '
                                     'hold a pen for normal '
                                     'handwriting.',
                                     wrapWidth=1.25
                                     )

        if isinstance(intro_text1, str):
            text1_stim.setText(intro_text1)
        elif isinstance(intro_text1, visual.TextStim):
            text1_stim = intro_text1
        intro_graphics['text1'] = text1_stim

        # Intro Text part 2
        text2_stim = visual.TextStim(self.win, units='norm',
                                     pos=(0, -0.2),
                                     height=0.066,
                                     color='green',
                                     text='Press the pen on the above '
                                     'target to start the '
                                     'validation, or the ESC key '
                                     'to skip the procedure.')
        if isinstance(intro_text2, str):
            text2_stim.setText(intro_text2)
        elif isinstance(intro_text2, visual.TextStim):
            text2_stim = intro_text2
        intro_graphics['text2'] = text2_stim

        self._penStim = None
        if self._displayPenPosition:
            # Validation Screen Graphics
            self._penStim = visual.Circle(self.win,
                                          radius=4,
                                          fillColor=[255, 0, 0],
                                          lineColor=[255, 0, 0],
                                          lineWidth=0,
                                          edges=8,  # int(np.pi*radius),
                                          units='pix',
                                          colorSpace='rgb255',
                                          opacity=0.9,
                                          contrast=1,
                                          interpolate=True,
                                          autoLog=False)

        if self._targetStim is None:
            self._targetStim = TargetStim(win,
                                          radius=16,
                                          fillcolor=[64, 64, 64],
                                          edgecolor=[192, 192, 192],
                                          edgewidth=1,
                                          dotcolor=[255, 255, 255],
                                          dotradius=3,
                                          units='pix',
                                          colorspace='rgb255',
                                          opacity=1.0,
                                          contrast=1.0
                                          )
        if intro_target_pos:
            self._targetStim.setPos(intro_target_pos)

        intro_graphics['target'] = self._targetStim

        if self._positionGrid is None:
            self._positionGrid = PositionGrid(
                winSize=win.monitor.getSizePix(),
                shape=[
                    3,
                    3],
                scale=0.9,
                posList=None,
                noiseStd=None,
                firstposindex=0,
                repeatfirstpos=True)

        # IntroScreen Graphics
        finished_graphics = self._finsihedScreenGraphics = OrderedDict()

        finished_graphics['title'] = visual.TextStim(
            self.win, units='norm', pos=(
                0, .9), height=0.1, text='Validation Complete')
        finished_graphics['result_status'] = visual.TextStim(
            self.win, units='norm', pos=(
                0, .7), height=0.07, color='blue', text='Result: {}')
        finished_graphics['result_stats'] = visual.TextStim(self.win, units='norm', pos=(
            0, .6), height=0.05, text='{}/{} Points Validated. Min, Max, Mean Errors: {}, {}, {}')
        finished_graphics['exit_text'] = visual.TextStim(
            self.win, units='norm', pos=(
                0, .5), height=0.05, text='Press any key to continue...')

    @property
    def targetStim(self):
        return self._targetStim

    @targetStim.setter
    def targetStim(self, ts):
        self._targetStim = ts

    @property
    def positionGrid(self):
        return self._positionGrid

    @positionGrid.setter
    def positionGrid(self, ts):
        self._positionGrid = ts

    def _enterIntroScreen(self):
        kb = self.io.devices.keyboard
        pen = self.io.devices.tablet

        exit_screen = False
        hitcount = 0
        pen.reporting = True
        kb.getPresses()

        while exit_screen is False:
            for ig in self._introScreenGraphics.values():
                ig.draw()
            samples = pen.getSamples()
            if samples:
                self._drawPenStim(samples[-1])
                spos = samples[-1].getPixPos(self.win)
                if samples[-1].pressure > 0 and \
                        self._introScreenGraphics['target'].contains(spos):
                    if hitcount > 10:
                        exit_screen = True
                    hitcount = hitcount + 1
                else:
                    hitcount = 0
            self.win.flip()
            if 'escape' in kb.getPresses():
                exit_screen = True
                pen.reporting = False
                break

        pen.reporting = False
        return True

    def _enterValidationSequence(self):
        val_results = dict(target_data=dict(), avg_err=0, min_err=1000,
                           max_err=-1000, status='PASSED', point_count=0,
                           ok_point_count=0)

        self._lastPenSample = None

        #kb = self.io.devices.keyboard
        pen = self.io.devices.pen

        self._positionGrid.randomize()

        pen.reporting = True
        for tp in self._positionGrid:
            self._targetStim.setPos(tp)
            self._targetStim.draw()
            targ_onset_time = self.win.flip()

            pen.clearEvents()

            val_sample_list = []

            while len(val_sample_list) < self.NUM_VALID_SAMPLES_PER_TARG:
                if core.getTime() - targ_onset_time > self.TARGET_TIMEOUT:
                    break
                self._targetStim.draw()

                samples = pen.getSamples()
                for s in samples:
                    spos = s.getPixPos(self.win)
                    if s.pressure > 0 and self.targetStim.contains(spos):
                        dx = math.fabs(tp[0] - spos[0])
                        dy = math.fabs(tp[1] - spos[1])
                        perr = math.sqrt(dx * dx + dy * dy)
                        val_sample_list.append((spos[0], spos[1], perr))
                    else:
                        val_sample_list = []

                if samples:
                    self._drawPenStim(samples[-1])
                    self._lastPenSample = samples[-1]
                elif self._lastPenSample:
                    self._drawPenStim(self._lastPenSample)
                self.win.flip()

            tp = int(tp[0]), int(tp[1])
            val_results['target_data'][tp] = None
            val_results['point_count'] = val_results['point_count'] + 1

            if val_sample_list:
                pos_acc_array = np.asarray(val_sample_list)
                serr_array = pos_acc_array[:, 2]

                targ_err_stats = val_results['target_data'][tp] = dict()
                targ_err_stats['samples'] = pos_acc_array
                targ_err_stats['count'] = len(val_sample_list)
                targ_err_stats['min'] = serr_array.min()
                targ_err_stats['max'] = serr_array.max()
                targ_err_stats['mean'] = serr_array.mean()
                targ_err_stats['median'] = np.median(serr_array)
                targ_err_stats['stdev'] = serr_array.std()

                val_results['min_err'] = min(
                    val_results['min_err'], targ_err_stats['min'])
                val_results['max_err'] = max(
                    val_results['max_err'], targ_err_stats['max'])

                val_results['avg_err'] = val_results[
                    'avg_err'] + targ_err_stats['mean']
                val_results['ok_point_count'] = val_results[
                    'ok_point_count'] + 1
            else:
                val_results['status'] = 'FAILED'

            self._lastPenSample = None

        if val_results['ok_point_count'] > 0:
            val_results['avg_err'] = val_results[
                'avg_err'] / val_results['ok_point_count']

        pen.reporting = False

        return val_results

    def _enterFinishedScreen(self, results):
        kb = self.io.devices.keyboard

        status = results['status']
        ok_point_count = results['ok_point_count']
        min_err = results['min_err']
        max_err = results['max_err']
        avg_err = results['avg_err']
        point_count = results['point_count']
        self._finsihedScreenGraphics['result_status'].setText(
            'Result: {}'.format(status))

        self._finsihedScreenGraphics['result_stats'].setText(
            '%d/%d '
            'Points Validated.'
            'Min, Max, Mean '
            'Errors: '
            '%.3f, %.3f, %.3f'
            '' %
            (ok_point_count, point_count, min_err, max_err, avg_err))
        for ig in self._finsihedScreenGraphics.values():
            ig.draw()
        self.win.flip()
        kb.clearEvents()

        while not kb.getPresses():
            for ig in self._finsihedScreenGraphics.values():
                ig.draw()
            self.win.flip()

    def _drawPenStim(self, s):
        if self._displayPenPosition:
            spos = s.getPixPos(self.win)
            if spos:
                self._penStim.setPos(spos)
                if s.pressure == 0:
                    self._penStim.setFillColor([255, 0, 0])
                    self._penStim.setLineColor([255, 0, 0])
                else:
                    self._penStim.setFillColor([0, 0, 255])
                    self._penStim.setLineColor([0, 0, 255])

                self._penStim.draw()

    def run(self):
        """Starts the validation process. This function will not return
        until the validation is complete. The validation results are
        returned in dict format.

        :return: dist containing validation results.

        """

        continue_val = self._enterIntroScreen()

        if continue_val is False:
            return None

        # delay about 0.5 sec before staring validation
        ftime = self.win.flip()
        while core.getTime() - ftime < 0.5:
            self.win.flip()
            self.io.clearEvents()

        val_results = self._enterValidationSequence()

        # delay about 0.5 sec before showing validation end screen
        ftime = self.win.flip()
        while core.getTime() - ftime < 0.5:
            self.win.flip()
            self.io.clearEvents()

        self._enterFinishedScreen(val_results)
        self.io.clearEvents()
        self.win.flip()

        return val_results

        # returning None indicates to experiment that the vaidation process
        # was terminated by the user
        # return None

    def free(self):
        self.win = None
        self.io = None
        self._finsihedScreenGraphics.clear()
        self._introScreenGraphics.clear()
        self._targetStim = None
        self._penStim = None

    def __del__(self):
        self.free()



from __future__ import print_function
from builtins import object
import pytest
import shutil
from tempfile import mkdtemp
import os


"""Each test class creates a context subclasses _baseVisualTest to run a series
of tests on a single graphics context (e.g. pyglet with shaders)

To add a new stimulus test use _base so that it gets tested in all contexts

"""
import psychopy
from psychopy import experiment
import psychopy.scripts.psyexpCompile as psyexpCompile

import codecs
from pathlib import Path

keepFiles = False

thisDir = Path(__file__).parent
psychoRoot = Path(psychopy.__file__).parent
demosDir = psychoRoot / 'demos'
testsDataDir = psychoRoot/'tests/data'



class Test_PsychoJS_from_Builder(object):
    """Some tests just for the window - we don't really care about what's drawn inside it
    """
    @pytest.mark.usefixtures("get_app")
    def setup_class(self, get_app):
        if keepFiles:
            self.temp_dir = Path.home() / "Desktop" / "tmp"
        else:
            self.temp_dir = Path(mkdtemp(prefix='psychopy-test_psychojs'))
        self.builderView = get_app().newBuilderFrame()  # self._app comes from requires_app

    def teardown_class(self):
        if not keepFiles:
            shutil.rmtree(self.temp_dir)

    def writeScript(self, exp, outFolder):
        script = exp.writeScript(expPath=outFolder, target="PsychoJS")
        with codecs.open(outFolder/'index.html', 'w',
                         encoding='utf-8-sig') as f:
            f.write(script)

    def compileScript(self, infile=None, version=None, outfile=None):
        """
        Compile script used to test whether JS modular files are written
        :param infile: psyexp file
        :param version: Version to use
        :param outfile: For testing JS filename
        :return: True
        """
        psyexpCompile.compileScript(infile, version, outfile)
        return True

    def test_stroop(self):
        #load experiment
        exp = experiment.Experiment()
        exp.loadFromXML(demosDir/'builder'/'Experiments'/'stroop'/'stroop.psyexp')
        # try once packaging up the js libs
        exp.settings.params['JS libs'].val = 'remote'
        outFolder = self.temp_dir/'stroopJS_remote/html'
        os.makedirs(outFolder)
        self.writeScript(exp, outFolder)

    def test_blocked(self):
        # load experiment
        exp = experiment.Experiment()
        exp.loadFromXML(demosDir/'builder'/'Design Templates'/'randomisedBlocks'/'randomisedBlocks.psyexp')
        # try once packaging up the js libs
        exp.settings.params['JS libs'].val = 'packaged'
        outFolder = self.temp_dir/'blocked_packaged/html'
        os.makedirs(outFolder)
        self.writeScript(exp, outFolder)
        print("files in {}".format(outFolder))

    def test_JS_script_output(self):
        # Load experiment
        exp = experiment.Experiment()
        exp.loadFromXML(demosDir/'builder'/'Experiments'/'stroop'/'stroop.psyexp')
        outFolder = self.temp_dir/'stroopJS_output/html'
        outFile = outFolder/'stroop.js'
        os.makedirs(outFolder)
        # Compile scripts
        assert(self.compileScript(infile=exp, version=None,
                                  outfile=str(outFile)))
        # Test whether files are written
        assert(os.path.isfile(os.path.join(outFolder, 'stroop.js')))
        assert(os.path.isfile(os.path.join(outFolder, 'stroop-legacy-browsers.js')))
        assert(os.path.isfile(os.path.join(outFolder, 'index.html')))
        assert(os.path.isdir(os.path.join(outFolder, 'resources')))

    def test_getHtmlPath(self):
        """Test retrieval of html path"""
        self.temp_dir = mkdtemp(prefix='test')
        fileName = os.path.join(self.temp_dir, 'testFile.psyexp')
        htmlPath = os.path.join(self.temp_dir, self.builderView.exp.htmlFolder)
        assert self.builderView._getHtmlPath(fileName) == htmlPath

    def test_getExportPref(self):
        """Test default export preferences"""
        assert self.builderView._getExportPref('on Sync')
        assert not self.builderView._getExportPref('on Save')
        assert not self.builderView._getExportPref('manually')
        with pytest.raises(ValueError):
            self.builderView._getExportPref('DoesNotExist')

    def test_onlineExtraResources(self):
        """Open an experiment with resources in the format of 2020.5
        (i.e. broken with \\ and with .. at start)"""
        expFile = (testsDataDir /
                  'broken2020_2_5_resources/broken_resources.psyexp')
        exp = experiment.Experiment()
        exp.loadFromXML(expFile)
        resList = exp.settings.params['Resources'].val
        print(resList)
        assert type(resList) == list
        assert (not resList[0].startswith('..'))


if __name__ == '__main__':
    cls = Test_PsychoJS_from_Builder()
    cls.setup_class()
    cls.test_stroop()
    cls.test_blocked()
    cls.test_JS_script_output()
    cls.test_onlineExtraResources()
    cls.teardown_class()
